/* automatique.c
  genere automatiquement un modele de chose a taper.
  
  The layout for the dvorak keyboard is deduced of a photography in
  Ben Shneiderman, _Designing the User Interface_, Addison Wesley, Reading,
  Massachussets, 1987, p. 230.
  He cites K. H. E. Kroemer, "Human engineering the keyboard", _Human Factors_,
  V. 14, N. 1, pp. 51--63, Fevrier 1972 et A. Martin, "A new keyboard layout",
  _Applied Ergonomics", V. 3, N. 1, 1972.

  jm, avril 1991
  jm, 22 mai 1991 : corrige le bug stupide de remodeler qui bloquait sur les
  	deux premiers caracteres.
  jm, 27 mai 1991 : maintenant on teste le debordement du tableau vues.
 */

# include <stdio.h>
# include "inline.h"
# include <ctype.h>

extern int erreurmax, okmax;	/* dans options.c */


/*
 Un ordre dans lequel on peut apprendre les touches sur un clavier
 et les touches avec lesquelles introduire les nouvelles touches
 */
struct apprendre {
  char * ordre;
# ifdef OTER
  char * avec;
  char prems[2];
  char * sequence;
# endif
} apprendre[] = {
				/* pour un clavier QWERY */
  { ("fjdksla;rueiwoqpvmc,x.z/ghytnb\
4738291056\
FJDKSLA:RUEIWOQPVMC<X>Z?GHYTNB\
!@#$&*()%^")
# ifdef OTER
     ,
      ("fjfjfjfjfjdksla;fjdksla;\
fjfjfjfjdksla;fj\
FFFFJJJJFDSAJKL;FDSA\
JKL;FJFJFJFDSAJKL:FJ")
	,
	{ 'f', 'd' },
	"ffff dddd fdfd dfdf"
# endif
	},
				/* pour un clavier dvorak */
	  {
	  ("uhetonaspg.c,r'lkmjwqv;z\
idyfxb4738291056\
UHETONASPG>C<R\"LKMJWQW:Z\
IDYFXB$&#*@(!)%%^")
# ifdef OTER
	    ,
	    ("huuhuhuhuhetonaauhetonaa\
uhuhuhuhetonaauh\
UUUUHHHHUEOAHTNSUEOAHTNS\
UHUHUHUEOAHTNSUH"),
	     { 'u', 'e' },
	   "uuuu eeee eueu ueue"
# endif
	},
};

int iordre = -1;

/*
 Parametres de l'apprentissage automatique, importes de xtype.c
 */
extern int quoi;		/* 0 : querty, 1 : dvorak */
extern int nlignes;		/* nbre de lignes ds une sequence */
extern int ngroupe;		/* nbre de groupes ds une sequence */
extern int ncars;		/* nbre de cars ds une sequence */
  
extern int pour_nouvelle;	/* nombre de sequences sans retirer de touches
				   avant de rajouter une nouvelle touche */

extern int dudebut;		/* si on prend la nouvelle touche en partant
				   du debut ; sinon on tourne dans la liste
				   des touches a apprendre */

# ifdef OTER
/*
 Locales
 */

char dvues[256];		/* la touche i vaut 1 si elle est vue */
int nvues = 0;			/* nombre de touches vues */

/*
 Exporte pour ecrire()
 */
char vues[256];			/* les touches vues : chaine de cars */


/* remodeler  --  renvoie une nouvelle chaine modele */
char * 
remodeler(cava, cavapas)
     char * cava, * cavapas;
{
  static int sansoter;
  static char * texte;
  char * p;
  int i, l;


				/* premiere invocation */
  if (texte == 0){
				/* allocation */
    texte = xmalloc(80 + nlignes * (ncars + 1) * ngroupe + 1);
				/* memoire */
    if (cava != NULL)
      while(*cava && nvues < sizeof vues - 1){
	if (*cava > ' '){
	  dvues[*cava] = 1;
	  vues[nvues++] = *cava;
	  jesais(*cava);
	}
	cava += 1;
      }
  }
  p = texte;
				/* oter les merdiques */
  if (cavapas != NULL && strlen(cavapas) > 0){
    while(*cavapas){
      dvues[*cavapas] = 0;
      jesaispas(*cavapas++);
    }    
    for(i = nvues = 0; i < sizeof dvues; i += 1)
      if (dvues[i] != 0)
	vues[nvues++] = i;
    sansoter = 0;
  } else
    sansoter += 1;

  if (nvues < 1){		/* il faut en connaitre */
    dvues[apprendre[quoi].prems[0]]
      = dvues[apprendre[quoi].prems[1]]
	= 1;
    vues[0] = apprendre[quoi].prems[0];
    vues[1] = apprendre[quoi].prems[1];
    jesais(vues[0]);
    jesais(vues[1]);
    nvues = 2;
    strcpy(p, apprendre[quoi].sequence);
    p += strlen(p);
    *p++ = '\n';
    l = 1;
  } else if (sansoter == pour_nouvelle){
    if (nvues != strlen(apprendre[quoi].ordre)){
      char nouvelle;		/* une nouvelle touche*/
    
      if (dudebut == 1)
	iordre = -1;
      sansoter = 0;
      while(dvues[apprendre[quoi].ordre[++iordre]] != 0
	    || apprendre[quoi].ordre[iordre] == 0)
	if (apprendre[quoi].ordre[iordre] == 0)
	  iordre = 0;
      
      nouvelle = apprendre[quoi].ordre[iordre];
      dvues[nouvelle] = 1;
      vues[nvues++] = nouvelle;
      jesais(nouvelle);

      for(i = 0; i < 4; i += 1)
	*p++ = nouvelle;
      *p++ = ' ';
      
      for(i = 0; i < 2; i += 1){
	*p++ = nouvelle;
	*p++ = apprendre[quoi].avec[iordre];
      }
      *p++ = ' ';
      
      for(i = 0; i < 2; i += 1){
	*p++ = apprendre[quoi].avec[iordre];
	*p++ = nouvelle;
      }
      *p++ = '\n';
      l = 1;
    } else if (erreurmax > 0){
      erreurmax -= 1;
      l = 0;
    } else
      okmoymax /= 2;
  } else
    l = 0; 
  vues[nvues] = 0;
				/* le texte a taper */
  for(; l < nlignes; l += 1){
    int j;
    for(j = 0; j < ngroupe; j += 1){
      for(i = 0; i < ncars; i += 1)
	*p++ = vues[(rand() / 7) % nvues];
      *p++ = ' ';
    }
    p[-1] = '\n';
  }
  
  *p = 0;
  return texte;
}
# else /* OTER */

# define BEAUCOUP 1024

char vues[BEAUCOUP];		/* l'urne ou se font les tirages aleatoires */
int nvues;

/* remodeler  --  faire un nouveau modele */
char *
remodeler(cava, cavapas)
     char * cava, * cavapas;
{
  static char * texte;
  static int sansoter;

  if (texte == 0){
    /*
      premiere fois : allouer le texte, initialiser vues.
     */
    texte = (char *)xmalloc(80 + nlignes * (ncars + 1) * ngroupe + 1);
    if (cava != NULL && cava[0] != 0){
      /* trier ca va, le mettre dans vues */
      int i, j, t;
      for(i = 1; cava[i] != 0; i += 1){
	t = cava[i];
	for(j = i; j > 0 && cava[j - 1] > t; j -= 1)
	  cava[j] = cava[j - 1];
	cava[j] = t;
      }

      for(; *cava && nvues < sizeof vues - 1; cava += 1)
	if (isgraph(*cava)){
	  vues[nvues++] = *cava;
	  jesais(*cava);
	}
    } else {
      /* cava est vide, initialiser vues avec apprendre */
      vues[0] = apprendre[quoi].ordre[0];
      vues[1] = apprendre[quoi].ordre[1];
      if (vues[1] < vues[0]){
	vues[0] ^= vues[1];
	vues[1] ^= vues[0];
	vues[0] ^= vues[1];
      }
      jesais(vues[0]);
      jesais(vues[1]);
      vues[2] = 0;
      nvues = 2;
    }
  } else {
    char * t, * nextchar();

    /*
      les autres fois, envisager d'ajouter, mettre l'urne a jour
     */
    if (cavapas == 0 || strlen(cavapas) == 0)
      sansoter += 1;
    else
      sansoter = 0;
    
    if (sansoter == pour_nouvelle){
      sansoter = 0;
      if ((t = nextchar()) == 0){
	/* il n'y a plus de caracteres a ajouter,
	   diminuer les erreurs ou le temps
	   */
	if (erreurmax > 0)
	  erreurmax -= 1;
	else if (okmax > 0)
	  okmax /= 2;
      }
    } else
      t = NULL;
    majurne(cava, cavapas, t);
  }

  { register char * p = texte;
    int i, j, l;
    
    for(l = 0; l < nlignes; l += 1){
      for(j = 0; j < ngroupe; j += 1){
	for(i = 0; i < ncars; i += 1)
	  *p++ = vues[(rand() / 7) % nvues];
	*p++ = ' ';
      }
      p[-1] = '\n';
    }
  }
  return texte;
}

majurne(ok, pasok, ajouter)
     char * ok, * pasok, * ajouter;
{
  int frequence[256];
  int i, j;
  int somme, max;
  extern int detail;		/* dans options.c */
  
  for(max = somme = i = j = 0; i < 256; i += 1){
    for(frequence[i] = 0; vues[j] == i; j += 1)
      frequence[i] += 1;
    somme += frequence[i];
    if (frequence[i] > max)
      max = frequence[i];
  }

  if (detail > 0)
    fprintf(stderr, "ok : \"%s\", bad : \"%s\", add \"%s\", somme %d, max %d\n",
	    ok, pasok, ajouter, somme, max);

  /* diviser par deux le nombre d'occurences des caracteres qui vont */
  if (ok != 0)
    for(;*ok; ok++)
      if (frequence[*ok] > 1){
	somme -= frequence[*ok];
	frequence[*ok] /= 2;
	somme += frequence[*ok];
      }

  /* multiplier par deux le nombre d'occurences des caracteres qui vont */
  if (pasok != 0)
    for(; *pasok; pasok++)
      if (frequence[*pasok] < somme / 4){
	somme += frequence[*pasok];
	frequence[*pasok] *= 2;
      }

  /* si il y a un caractere trop mal connu, refuser d'ajouter */
  if (somme > 4 && max > somme / 4)
    ajouter = 0;

  /* ajouter les nouveaux caracteres, avec une frequence maximale */
  if (ajouter != 0){
    for(; *ajouter; ajouter++){
      frequence[*ajouter] = max;
      somme += max;
      jesais(*ajouter);
    }
  }

  /* dimimuer toutes les frequences > a un pour que ca rentre dans l'urne */
  while(somme > BEAUCOUP){
    somme = 0;
    for(i = 0; i < 256; i += 1){
      if (frequence[i] > 1)
	frequence[i] /= 2;
      somme += frequence[i];
    }
  }

  for(i = nvues = 0; i < 256; i += 1)
    for(j = frequence[i]; j > 0; j -= 1)
      vues[nvues++] = i;
  vues[nvues] = 0;
}

static estdans(c, s)
     register char * s;
     register c;
{
  while(*s != 0 && *s != c)
    s += 1;
  return *s == c;
}

char *
nextchar()
{
  static char t[2];
  int i;

  if (dudebut){
    for(i = 0; apprendre[quoi].ordre[i] != 0; i += 1)
      if (! estdans(apprendre[quoi].ordre[i], vues)){
	t[0] = apprendre[quoi].ordre[i];
	t[1] = 0;
	return t;
      }
  } else {
    static index = 0;
    static ln = 0;
    
    if (ln == 0)
      ln = strlen(apprendre[quoi].ordre);
    
    i = index;
    do
      if (! estdans(apprendre[quoi].ordre[i], vues)){
	index = (i + 1) % ln;
	t[0] = apprendre[quoi].ordre[i];
	t[1] = 0;
	return t;
      } else
	i = (i + 1) % ln;
    while(i != index);
  }
  return 0;
}
# endif /* OTER */
