/* clavier.c
 A partir d'une description de la taille des touches et des keycodes
 qu'elles generent, fabriquer une representation du clavier affichable

 jm, avril 1991
 
 jm, 21 mai 1991 : ajoute la description du clavier de l'apollo
 jm, 27 mai 1991 : ajoute les claviers cables qwerty et cable, sorti
                   la remise a zero des flags des touches de la section
		   ou on place.
 jm, 14 juin 1991 : ajoute la definition du clavier NCD, decale les
 		  touches des qwerty et dvorak cables.
 */

# include <stdio.h>
void * malloc();
# include <ctype.h>
# include <X11/Xlib.h>

# include "clavier.h"
# include "menu.h"

/*
 Importes
 */
extern int ispace;		/* espace entre le cadre et les chars */
extern int tcadre;		/* epaisseur du cadre */
extern int vspace, hspace;	/* espaces entre les cadres */
extern Display * d;
extern Window w;

/*
 Exportes
 */
struct touche * touche, trop;
struct touche * ascii2touche[1 << (sizeof(char) * 8)];

int rsize;			/* la taille (interieure) d'une touche */

# include "inline.h"

     /*
       Pour definir un nouveau clavier :

       Les touches sont organisees en tableaux de lignes ;
       chaque touche est representee par une structure cab.
       Normalement le champs k contient le keycode, le champs largeur contient
       la proportion de la largeur par rapport a la hauteur.

       Si le flag ne contient pas le bit YALKEYCODE, le programme prend le
       keycode de la touche precedente plus un.

       Si la largeur vaut 0, le programme fabrique une touche qui s'etend
       jusqu'a la droite du clavier ; ca ne peut pas marcher sur la premiere
       ligne, ni ailleurs que sur la derniere touche d'une ligne.

       Si le flag contient le bit SOULIGNER, la touche est entouree avec un
       trait gras (pour montrer ou poser les indexs).

       Si le flag contient le bit PLACER, la touche sera espacee de coordx
       fois la taille d'une touche avec la precedente (pour les claviers pas
       rectangulaires).

       Si le flag contient le bit CESTFINI, le cab sert seulement a
       marquer la fin de la ligne.

      On peut trouver les keycodes qui correspondent aux touches avec
      la commande xmodmap -pk

        Une fois les lignes definies, il faut (1) ajouter une ligne dans
	cab qui pointe sur les lignes, (2) ajouter le code du clavier dans
	clavier.h (3) definir une nouvelle entree dans keyboard_type.
      */
# include "cab.h"

/* incrementer le keycode, touche carree */
# define X	{ 0, 0, 1. }
/* idem pour poser les indexs */
# define S	{ SOULIGNER, 0, 1. }
/* fin de ligne de touches */
# define Y	{ CESTFINI }

/*
 Voila la definition (de la partie interressante) des claviers
 */
static struct cab
  /*
    Qwerty et dvorak cable (label est au courant)
   */
  cable_ligne1[] = { { YALKEYCODE, 1, 1. }, X, X, X, X, X, X, X, X, X, Y },
  cable_ligne2[] = { { PLACER, 0, 1., .3 }, X, X, X, X, X, X, X, X, X, Y },
  cable_ligne3[] = { { PLACER, 0, 1., .6 }, X, X, S, X, X, S, X, X, X, Y },
  cable_ligne4[] = { { PLACER, 0, 1., .9 }, X, X, X, X, X, X, X, X, X, Y },
  /*
    SUN avec X11 R4 athena
   */
  sun_ligne1[] = { {YALKEYCODE, 36, 1.}, /* escape */
		 X, X, X, X, X, X, X, X, X, X, X, X,
		 {YALKEYCODE, 95, 1.}, /* contre-oblique */
		 {YALKEYCODE, 49, 1.}, /* contre-apostrophe */
		 Y},
  sun_ligne2[] = { {YALKEYCODE, 60, 1.3}, /* tab */
		 X, X, X, X, X, X, X, X, X, X, X, X,
		 { 0, 0, 0. }, Y},
  sun_ligne3[] = { {YALKEYCODE, 83, 1.6}, /* control */
		 X, X, X, S, X, X, S, X, X, X, X,
		 {YALKEYCODE, 96, 0.}, /* return */
		 Y},
  sun_ligne4[] = { {YALKEYCODE, 106, 2.1}, /* shift */
		 X, X, X, X, X, X, X, X, X, X,
		 {0, 0, 2.}, {YALKEYCODE, 118, 0.}, Y}, /* shift, line feed */
  /*
    SONY avec sonyOS
    */
  sony_ligne1[] = { {YALKEYCODE, 18, 1.},
		 X, X, X, X, X, X, X, X, X, X, X, X, X,
		{ 0, 0, 1.5}, Y},
  sony_ligne2[] = { {YALKEYCODE, 33, 1.3},
		 X, X, X, X, X, X, X, X, X, X, X, X, {0 },  Y},
  sony_ligne3[] = { {YALKEYCODE, 47, 1.6},
		 X, X, X, S, X, X, S, X, X, X, X, X,{0, 0, 0}, Y},
  sony_ligne4[] = { {YALKEYCODE, 61, 2.1},
		 X, X, X, X, X, X, X, X, X, X, X,
		 {0, 0, 0}, Y},
  /*
    apollo avec domain 10.2
    */
  apollo_ligne1[] = {
    { YALKEYCODE|PLACER, 128, 1., .8 }, 
    { YALKEYCODE, 241, 1. },
    X, X, X, X, X, X, X, X, X, X, X, X,
    { YALKEYCODE, 190, 1.5},
    Y },
  apollo_ligne2[] = {
    { YALKEYCODE|PLACER, 188, 1.5, .8 },
    { YALKEYCODE, 225, 1. },
    X, X, X, X, X, X, X, X, X, X, X,
    { YALKEYCODE, 191, 0.},
    Y },
  apollo_ligne3[] = {
    { YALKEYCODE, 186, 1. },	/* control */
    { YALKEYCODE, 185, 1.5 }, /* shift */
    { YALKEYCODE, 209, 1. }, X, X, S, X, X, S, X, X, X, X,
    { YALKEYCODE, 189, 2. }, /* return */
    { YALKEYCODE, 222, 0. },
    Y },
  apollo_ligne4[] = {
    { YALKEYCODE, 187, 1. },	/* repeat */
    { YALKEYCODE, 183, 2.}, /* shift */
    { YALKEYCODE, 193, 1. }, X, X, X, X, X, X, X, X, X, X,
    { YALKEYCODE, 184, 0. }, /* shift */
    /* pop */
    Y },
  /*
    NCD
   */
  ncd_ligne1[] = { {YALKEYCODE, 8, 1.}, /* escape */
		     { YALKEYCODE, 22, 1.}, /* 1 */
		     { YALKEYCODE, 30, 1. }, /* 2 */   
		     { YALKEYCODE, 38, 1. }, /* 3 */   
		     { YALKEYCODE, 37, 1. }, /* 4 */   
		     { YALKEYCODE, 46, 1. }, /* 5 */   
		     { YALKEYCODE, 54, 1. }, /* 6 */   
		     { YALKEYCODE, 61, 1. }, /* 7 */   
		     X,	/* 8 */
		     { YALKEYCODE, 70, 1. }, /* 9 */   
		     { YALKEYCODE, 69, 1. }, /* 0 */   
		     { YALKEYCODE, 78, 1. }, /* - */   
		     { YALKEYCODE, 85, 1. }, /* + */   
		     { YALKEYCODE, 14, 1. }, /* ` */   
		     { YALKEYCODE, 102, 1.7 }, /* backspace */   
		     Y},
  ncd_ligne2[] = { {YALKEYCODE, 13, 1.3}, /* tab */
		     { YALKEYCODE, 21, 1.}, /* q */
		     { YALKEYCODE, 29, 1. }, /* w */   
		     { YALKEYCODE, 36, 1. }, /* e */   
		     { YALKEYCODE, 45, 1. }, /* r */   
		     { YALKEYCODE, 44, 1. }, /* t */   
		     { YALKEYCODE, 53, 1. }, /* y */   
		     { YALKEYCODE, 60, 1. }, /* u */   
		     { YALKEYCODE, 67, 1. }, /* i */
		     { YALKEYCODE, 68, 1. }, /* o */   
		     { YALKEYCODE, 77, 1. }, /* p */   
		     { YALKEYCODE, 84, 1. }, /* [ */   
		     { YALKEYCODE, 91, 1. }, /* ] */   
		     { YALKEYCODE, 92, 1. }, /* contreoblique */   
		     { YALKEYCODE, 100, 0. }, /* del */   
		     Y},
  ncd_ligne3[] = { {YALKEYCODE, 17, 1.6}, /* control */
		     { YALKEYCODE, 28, 1.}, /* a */
		     { YALKEYCODE, 27, 1. }, /* s */   
		     { YALKEYCODE, 35, 1. }, /* d */   
		     { YALKEYCODE|SOULIGNER, 43, 1. }, /* f */   
		     { YALKEYCODE, 52, 1. }, /* g */   
		     { YALKEYCODE, 51, 1. }, /* h */   
		     { YALKEYCODE|SOULIGNER, 59, 1. }, /* j */   
		     { YALKEYCODE, 66, 1. }, /* k */
		     { YALKEYCODE, 75, 1. }, /* l */   
		     { YALKEYCODE, 76, 1. }, /* ; */   
		     { YALKEYCODE, 82, 1. }, /* apostrophe */   
		     { YALKEYCODE, 90, 2. }, /* return */   
		     { YALKEYCODE, 87, 0. }, /* line feed */   
		     Y},
  ncd_ligne4[] = { {YALKEYCODE, 18, 2.05}, /* shift */
		     { YALKEYCODE, 26, 1.}, /* z */
		     { YALKEYCODE, 34, 1. }, /* x */   
		     { YALKEYCODE, 33, 1. }, /* c */   
		     { YALKEYCODE, 42, 1. }, /* v */   
		     { YALKEYCODE, 50, 1. }, /* b */   
		     { YALKEYCODE, 49, 1. }, /* n */   
		     { YALKEYCODE, 58, 1. }, /* m */   
		     { YALKEYCODE, 65, 1. }, /* , */
		     { YALKEYCODE, 73, 1. }, /* . */   
		     { YALKEYCODE, 74, 1. }, /* / */   
		     { YALKEYCODE, 89, 1.5 }, /* shift */   
		     { YALKEYCODE, 99, 1. }, /* up */   
		     { YALKEYCODE, 103, 1. }, /* option */
		     Y},


  * cab[][5] = {
    { cable_ligne1, cable_ligne2, cable_ligne3, cable_ligne4, NULL },
    { cable_ligne1, cable_ligne2, cable_ligne3, cable_ligne4, NULL },
    { &sun_ligne1[0], &sun_ligne2[0], &sun_ligne3[0], &sun_ligne4[0], NULL },
    { &sony_ligne1[0], &sony_ligne2[0], &sony_ligne3[0],
	&sony_ligne4[0], NULL },
    { &apollo_ligne1[0], &apollo_ligne2[0], &apollo_ligne3[0],
	&apollo_ligne4[0], NULL },
    { &ncd_ligne1[0], &ncd_ligne2[0], &ncd_ligne3[0], &ncd_ligne4[0], NULL },
  };

# ifndef DEFAULT_KEYBOARD
#   define DEFAULT_KEYBOARD SUN
# endif /* DEFAULT_KEYBOARD */
int kbtype = DEFAULT_KEYBOARD;	/* index du clavier dans cab */

/*
 Voici le contenu du menu qui permet de choisir le type de clavier affiche
 */
struct entree
  keyboard_type[] = {
# ifndef ENGLISH
    { "qwerty cable", QWERTY_CABLE },
    { "dvorak cable",DVORAK_CABLE },
# else /* ENGLISH */
    { "wired qwerty", QWERTY_CABLE },
    { "wired dvorak", DVORAK_CABLE },
# endif /* ENGLISH */
    { "sun", SUN },
    { "sony", SONY},
    { "apollo", APOLLO },
    { "ncd unix", NCD },
    { NULL }
  };


# undef X
# undef Y
# undef S

/* clavier  --  construit la representation des touches */
clavier(fd, placer)
     XFontStruct * fd;
     int placer;		/* s'il faut positionner les touches */
{
  int i, k, l;
  int curx, cury;
  KeyCode kc;
  int * tmp, * sauve_flags();

  /* le cas echeant, raz */
  if (touche != 0){
    tmp = sauve_flags();
    for(i = 0; i < 256; i += 1)
      ascii2touche[i] = 0;
  } else {
    tmp = 0;
  }

  /* compter les touches, allouer, charger la table de traduction */
  { int ntouches;
    for(ntouches = l = 0; cab[kbtype][l] != NULL; l += 1){
      for(i = 0; cab[kbtype][l][i].flag != CESTFINI; i += 1)
	;
      ntouches += i;
    }
    if (placer){
      touche = (struct touche *)xmalloc((ntouches + 1) * sizeof touche[0]);
      rsize = 2 * (fd->ascent + fd->descent);
    }
    startKeymap(ntouches);
  }

  /* determiner la position, la taille et l'etiquette de chaque touche */
  if (placer)
    trop.x = cury = 0;
  for(k = l = 0; cab[kbtype][l] != NULL; l += 1){
    curx = 0;
    for(i = 0; cab[kbtype][l][i].flag != CESTFINI; i += 1){
      if (placer){
	touche[k].y = cury;	/* position */
	if ((cab[kbtype][l][i].flag & PLACER))
	  touche[k].x = (curx += rsize * cab[kbtype][l][i].coordx);
	else
	  touche[k].x = curx;

	if (cab[kbtype][l][i].largeur != 0) /* taille */
	  touche[k].width = rsize * cab[kbtype][l][i].largeur;
	else			/* la touche va jusqu'au bout */
	  touche[k].width = trop.x - (curx + tcadre + hspace + tcadre);
	
	touche[k].height = rsize;
	curx += tcadre + touche[k].width + tcadre + hspace;

	if (cab[kbtype][l][i].flag & SOULIGNER)
	  touche[k].flag |= SOULIGNER;
      }

      touche[k].flag &= ~GRISE;	/* flags */
      label(&touche[k],		/* etiquette */
	    ((cab[kbtype][l][i].flag&YALKEYCODE) ? kc = cab[kbtype][l][i].k : ++kc), fd);

      k += 1;
    }

    if (placer){
      if (curx > trop.x)
	trop.x = curx;
      cury += tcadre + rsize + tcadre + vspace;
      curx = 0;
    }
  }
  endKeymap();
  if (placer){
    trop.y = cury;
    touche[k].up = touche[k].down = NULL;
  } else
    restaure_flags(tmp);
}

/*
 etiquetter les touches
 */
static KeySym * keymap;
static int kmin, kmax, nsym_par_code;
static XKeyEvent e;
static char * etiquettes, * p;
  
startKeymap(n)
     int n;
{
  XDisplayKeycodes(d, &kmin, &kmax);
  if ((keymap = XGetKeyboardMapping(d, kmin, kmax - kmin + 1, &nsym_par_code))
      == NULL){
    fprintf(stderr,
# ifndef ENGLISH
	    "Incapable de charger la table de trad. du clavier, tant pis\n"
# else
	    "Unable to load the keyboard mapping table, quitting\n"
# endif
	    );
    exit(1);
  }

# define ASSEZ	8		/* pour les etiquettes, * nbre de touche */
  if (etiquettes != 0)
    free(etiquettes);
  etiquettes = p = (char *)xmalloc(n * ASSEZ);

  e.type = KeyPress;
  e.serial = 2;
  e.send_event = True;
  e.display = d;
  e.window = w;
  e.root = RootWindow(d, DefaultScreen(d));
  e.subwindow = w;
  e.time = CurrentTime;
  e.x = e.y = e.x_root = e.y_root = 0;
  e.same_screen = True;
}

endKeymap(){
  XFree((char *)keymap);
}

label(t, kc, fd)
     struct touche * t;
     KeyCode kc;
     XFontStruct * fd;
{
  char table[10];
  int combien;
  KeySym ks;

  if (kbtype == QWERTY_CABLE){
    t->up = p;
    *p ++ = " !@#$%^&*()QWERTYUIOPASDFGHJKL:ZXCVBNM<>?"[kc];
    *p ++ = 0;
    t->down = p;
    *p ++ = " 1234567890qwertyuiopasdfghjkl;zxcvbnm,./"[kc];
    *p ++ = 0;
    ascii2touche[t->up[0]] = ascii2touche[t->down[0]] = t;
  } else if (kbtype == DVORAK_CABLE){
    t->up = p;
    *p ++ = " !@#$%^&*()\"<>PYFGCRLAOEUIDHTNS:QJKXBMWVZ"[kc];
    *p ++ = 0;
    t->down = p;
    *p ++ = " 1234567890',.pyfgcrlaoeuidhtns;qjkxbmwvz"[kc];
    *p ++ = 0;
    ascii2touche[t->up[0]] = ascii2touche[t->down[0]] = t;
  } else /* demander au serveur */{
    t->up = t->down = "";

    e.state = 0;
    e.keycode = kc;

    combien = XLookupString(&e, table, sizeof table, &ks, NULL);
    if (combien == 1)
      ascii2touche[table[0]] = t;

    if (combien != 1 || !isgraph(table[0])){
      char * str;
      int n;

      ks = keymap[(kc - kmin) * nsym_par_code];

      if (ks == NoSymbol || (str = XKeysymToString(ks)) == NULL){
	t->up = "";
	t->down = "";
      } else {
	for(n = 0; (str[n] != 0 && str[n] != '_'
		    && XTextWidth(fd, str, n + 1) < t->width - 2 * ispace);
	    n += 1)
	  ;
	t->up = p;
	strncpy(p, str, n);
	p[n] = 0;
	p += n + 1;
      }
      return;
    }
  
    t->down = p;
    *p++ = table[0];
    *p++ = 0;
  
    e.state = ShiftMask;
    combien = XLookupString(&e, table, sizeof table, &ks, NULL);
    if (combien == 1)
      ascii2touche[table[0]] = t;

    if (combien != 1 || !isgraph(table[0]))
      t->up = "";
    else {
      t->up = p;
      *p++ = table[0];
      *p++ = 0;
    }
  }
}

int *
sauve_flags()
{
  int * t = (int *)xmalloc(256 * sizeof t[0]);
  int i;

  for(i = 0; i < 256; i += 1)
    if (ascii2touche[i] != 0 && (ascii2touche[i]->flag&GRISE))
      t[i] = GRISE;
    else
      t[i] = 0;
  return t;
}

restaure_flags(t)
     int * t;
{
  int i;

  for(i = 0; i < 256; i += 1)
    if (t[i] != 0 && ascii2touche[i] != 0)
      ascii2touche[i]->flag |= t[i];
  free(t);
}
