/*
 Gerer des menus (sans casccades)
 jm, 28 mai 1991
 */
# include <stdio.h>
# include <X11/Xlib.h>
# include "inline.h"
# include "menu.h"

extern Display * d;		/* commencer.c : dispay */
extern long noir, blanc;	/* commencer.c : foreground, background */

# define VESPACE	0	/* espace interne vertical */
# define HESPACE	2	/* espace interne horizontal */
# define CADRE		1	/* epaisseur des cadres */
# define OMBRE		2	/* ombre du menu */

/* une ligne de menu */
struct ligne {
  struct entree * e;		/* marque et code */
  int ncar;			/* combien de car. dans la marque */
  Window w;			/* quelle fenetre pour la ligne */
};

/* un menu */
struct menu {
  int width, height;		/* taille de la fenetre principale */
  int lheight, lwidth;		/* taille d'une ligne */
  int lx, ly;			/* ou on affiche dans la ligne */
  XFontStruct * police;		/* la police a utiliser */
  Window w;			/* la fenetre principale */
  int courante;			/* la ligne courante */
  struct ligne ligne[1];	/* necessairement le dernier champs */
};

GC gcinvert = 0, gcnormal;

/* trouver  --  transformer une fenetre en ligne dans le menu */
static struct ligne *
trouver(m, w)
     struct menu * m;
     register Window w;
{
  register struct ligne * l;

  for(l = &m->ligne[0]; l->e != 0; l += 1)
    if (w == l->w)
      return l;
  fprintf(stderr,
#	  ifndef ENGLISH
	  "choisir : trouver : impossible de trouver la fenetre %ld\n",
#	  else
	  "choisir: trouver: unable to find the window %ld\n"
#	  endif
	  w);
  return 0;
}

/* menu  --  fabriquer un menu */
struct menu *
menu(es, police)
     struct entree es[];
     XFontStruct * police;
{
  int nl, large;
  struct menu * m;

  /* Compte les lignes, mesure leur largeur */
  for(large = nl = 0; es[nl].marque != 0; nl += 1){
    int t = XTextWidth(police, es[nl].marque, strlen(es[nl].marque));
    if (t > large)
      large = t;
  }

  /* Alloue le menu et ses lignes, initialise les tailles */
  m = (struct menu *)xmalloc(sizeof m[0] + nl * sizeof m->ligne[0]);
  m->lx = HESPACE;
  m->ly = police->ascent + VESPACE;
  m->lwidth = large + 2 * HESPACE;
  m->lheight = police->ascent + police->descent + 2 * VESPACE;
  m->width = m->lwidth + 2 * CADRE + OMBRE;
  m->height = CADRE + nl * (m->lheight + CADRE) + OMBRE;
  m->courante = 0;

  /* Fabrique la premiere fenetre */
  { XSetWindowAttributes xswa;

    xswa.background_pixel = noir;
    xswa.save_under = True;
    xswa.override_redirect = True;
    m->w = XCreateWindow(d, RootWindow(d, DefaultScreen(d)),
			 0, 0, m->width, m->height, 0, /* geometrie, cadre */
			 CopyFromParent, InputOutput,  /* depth, class */
			 CopyFromParent, /* visual */
			 (CWBackPixel|CWSaveUnder|CWOverrideRedirect), &xswa);
  }

  /* Initialise les lignes et fabrique leurs fenetres */
  for(m->ligne[nl--].e = 0; nl >= 0; nl -= 1){
    m->ligne[nl].e = &es[nl];
    m->ligne[nl].ncar = strlen(es[nl].marque);
    m->ligne[nl].w = XCreateSimpleWindow(d, m->w,
					 0, nl * (m->lheight + CADRE),
					 m->lwidth, m->lheight,
					 CADRE, noir, blanc);
    XSelectInput(d, m->ligne[nl].w, (ButtonReleaseMask|ExposureMask
				    |EnterWindowMask|LeaveWindowMask));
  }
  XMapSubwindows(d, m->w);

  /* La premiere fois, fabrique les deux contextes graphiques */
  if (gcinvert == 0){
    XGCValues gcv;

    gcv.font = police->fid;
    gcv.foreground = noir;
    gcv.function = GXcopy;
    gcnormal = XCreateGC(d, m->w, GCFunction|GCForeground|GCFont, &gcv);

    gcv.plane_mask = noir^blanc;
    gcv.function = GXinvert;
    gcv.graphics_exposures = False;
    gcinvert = XCreateGC(d, m->w,
			 GCFunction|GCPlaneMask|GCGraphicsExposures, &gcv);
  }
  return m;
}

/* choisir  --  choisir une ligne dans un menu */
choisir(m, e)
     register struct menu * m;
     XButtonEvent * e;
{
  /* positionner et mapper la premiere fenetre */
  { register int oux, ouy;

    oux = e->x_root - 2;
    ouy = e->y_root;
    if (m->courante != ECHEC)
      ouy -= CADRE + m->courante * (m->lheight + CADRE) + m->lheight / 2;
  
    if (oux < 0)
      oux = 0;
    else if (oux + m->width > DisplayWidth(d, DefaultScreen(d)))

      oux = DisplayWidth(d, DefaultScreen(d)) - m->width;
    
    if (ouy < 0)
      ouy = 0;
    else if (ouy + m->height > DisplayHeight(d, DefaultScreen(d)))
      ouy = DisplayHeight(d, DefaultScreen(d)) - m->height;
    
    XMoveWindow(d, m->w, oux, ouy);
    XRaiseWindow(d, m->w);
    XMapWindow(d, m->w);
  }

  /* rouler sur les evenements */
  for(;;){
    register struct ligne * l;
    char * nomevent();
    XEvent e;

    XNextEvent(d, &e);
    switch(e.type){
# define when	break;case
    default:
      fprintf(stderr, 
	      "choisir : event %d (%s)\n", e.type, nomevent(e.type));

    when Expose:
      if ((l = trouver(m, e.xexpose.window)) != 0)
	XDrawString(d, l->w, gcnormal, m->lx, m->ly, l->e->marque, (l->ncar));

    when EnterNotify:
    case LeaveNotify:
      if ((l = trouver(m, e.xcrossing.window)) != 0){
	m->courante = (e.type == EnterNotify ? l - &m->ligne[0] : ECHEC);
	XCopyArea(d, l->w, l->w, gcinvert, 0, 0, m->lwidth, m->lheight, 0, 0);
      }

    when ButtonRelease:
      XUnmapWindow(d, m->w);
      return m->courante == ECHEC ? ECHEC : m->ligne[m->courante].e->code;
    }
  }
}
